package be.iminds.ilabt.jfed.highlevel.stitcher;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.lib.MockReplyLibrary;
import be.iminds.ilabt.jfed.lowlevel.lib.MockSfaConnection;
import be.iminds.ilabt.jfed.lowlevel.stitching.VlanRange;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.rspec_source.ImmutableRequestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.library.DataConversionUtils;
import be.iminds.ilabt.jfed.util.common.TextUtil;
import org.apache.xmlrpc.XmlRpcRequest;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * TestStitchingAMv2
 */
public class TestStitchingAMv3Server extends TestStitchingAMServer {
    private final static Logger LOG = LoggerFactory.getLogger(TestStitchingAMv3Server.class);

    private static enum State { NO, ALLOCATED, PROVISIONED, STARTING, READY, FAILED }
    private int statusRequestsReceived = 0;
    private int totalStatusRequestsReceived = 0;
    private final int statusRequestsBeforeReady;
    private State state = State.NO;

    private String manifest;

    private boolean answeredPermanentFailure = false;
    private boolean sawAnyRequest = false;
    private List<Integer> seenRequests = new ArrayList<>();

    public TestStitchingAMv3Server(String serverUrl, String cmUrn, String cmHostname,
                                   String stitchIface, VlanRange availableVlans, StitchingTestSetupConfig.StitchingTestSetupAmConfig amConfig,
                                   TestStitchingAMServer dependency) {
        super(serverUrl, cmUrn, cmHostname, stitchIface, availableVlans, amConfig, dependency);
        assert amConfig.hasV3;
        this.statusRequestsBeforeReady = 2;
    }

    public MockSfaConnection.MockXmlRpcAnswer matches(XmlRpcRequest request) {
        if (!MockReplyLibrary.getServerUrl(request).equals(serverUrl))
            return null;

        int callCount;
        if (request.getMethodName() != null)
            callCount = registerCall(request.getMethodName());
        else
            callCount = -1;

        if (amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_ALWAYS_TIMEOUT))
            return new MockSfaConnection.MockXmlRpcAnswer(new org.apache.http.conn.ConnectTimeoutException("Connect to server timed out"));

        if (request.getMethodName().equalsIgnoreCase("ListResources")) {
            assert request.getMethodName().equals("ListResources");
            assert request.getParameterCount() == 2;
            Map options = (Map) request.getParameter(1);
            assert !options.containsKey("geni_slice_urn");

            /*
             * struct ListResources(string credentials[], struct options)
             * Typical options, which we assume here:  (so no "geni_slice_urn")
             *   "geni_available": true, "geni_rspec_version": { "version": "3", "type": "geni" }, "geni_compressed": true
             * */
            if (!state.equals(State.NO)) throw new RuntimeException("am is in wrong state for "+request.getMethodName()+": "+state);

            boolean compressed = false;
            if (options.containsKey("geni_compressed")) {
                compressed = TextUtil.objectToBoolean(options.get("geni_compressed"));
            }
            assert compressed : "normally jFed sets compressed";

            assert advertisement != null;

            String resAdvertisement;
            if (compressed)
                resAdvertisement = DataConversionUtils.compressToBase64(advertisement);
            else
                resAdvertisement = advertisement;

            return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapSuccessStringReply(resAdvertisement));
        }
        if (request.getMethodName().equals("Allocate") && request.getParameterCount() >= 3) {
            if (!state.equals(State.NO)) throw new RuntimeException("TEST BUG DETECTED: am is in wrong state for "+request.getMethodName()+": "+state);

            if (answeredPermanentFailure)
                return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(5, "TEST BUG DETECTED: already replied with permanent failure. This retry shouldn't have been done."));

            LOG.debug("MOCK AMv3 Allocate callCount="+callCount+" hasSpecial="+amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_ONETIME_GENERIC_ALLOCATE));
            if (callCount == 1 && amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_ONETIME_GENERIC_ALLOCATE))
                return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(2, "Error reserving vlan tag for link0"));
            if (callCount == 1 && amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_ALLOCATE_ONE_TIMEOUT))
                return new MockSfaConnection.MockXmlRpcAnswer(new org.apache.http.conn.ConnectTimeoutException("Connect to server timed out"));
            if (callCount == 2 && amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_ALLOCATE_TWO_TIMEOUT))
                return new MockSfaConnection.MockXmlRpcAnswer(new org.apache.http.conn.ConnectTimeoutException("Connect to server timed out"));

            //struct Allocate(string slice_urn, struct credentials[], geni.rspec rspec, struct options)
            String requestRspec = (String) request.getParameter(2);

            if (!amConfig.noStitching) {
                String suggestedVlan = getSuggestedVlan(requestRspec);
                assert suggestedVlan != null;

                if (dependency != null && suggestedVlan.trim().equalsIgnoreCase("any")) {
                    String availableRangeStr = getAvailableVlans(new ImmutableRequestRspecSource(requestRspec, ModelRspecType.BASIC).getImmutableModelRspec());
                    if (availableRangeStr == null)
                        return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(5, "TEST BUG DETECTED: no available range"));
                    assert availableRangeStr != null;
                    VlanRange availableRange = new VlanRange(availableRangeStr);
                    if (availableRange.size() != 1 || !availableRange.contains(dependency.getActiveVlan())) //if available is just dependency.getActiveVlan(), it is allowed. otherwise not.
                        return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(5,
                                "TEST BUG DETECTED: a n y used while this server depends on another server. " +
                                        "availableRange=\""+availableRangeStr+"\" other server vlan: "+dependency.getActiveVlan()));
                }

                if (dependency != null && Integer.parseInt(suggestedVlan) != dependency.getActiveVlan())
                    return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(5,
                            "TEST BUG DETECTED: vlan "+suggestedVlan+" was requested while dependency has "+dependency.getActiveVlan()+" as active vlan"));

                if (!supportsAny && suggestedVlan.trim().equalsIgnoreCase("any")) {
                    if (sawAnyRequest)
                        return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(5, "TEST BUG DETECTED: not first request with vlan a n y"));
                    sawAnyRequest = true;

                    if (!authority.hasFlag(Server.Flag.featureStitchingAny))
                        return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(5, "TEST BUG DETECTED: request with vlan a n y for authority which is known not to support it"));

                    return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(2, "ERROR: vlan range any is invalid"));

                }

                //if any is supported or not, if suggested is not any, we need to check if it is available
                if (!suggestedVlan.equalsIgnoreCase("any")) {
                    int suggestedVlanInt = 0;
                    try {
                        suggestedVlanInt = Integer.parseInt(suggestedVlan);
                    } catch (NumberFormatException e) {
                        return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(5, "TEST BUG DETECTED: invalid vlan in suggestedVlan: \"" + suggestedVlan + "\""));
                    }
                    if (!availableVlans.contains(suggestedVlanInt) || amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_NO_VLANS)) {

                        if (seenRequests.contains(suggestedVlanInt))
                            return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(5, "TEST BUG DETECTED: not first request with vlan \"" + suggestedVlan + "\""));
                        seenRequests.add(suggestedVlanInt);

                        return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(2, "vlan tag " + suggestedVlanInt + " for " + cmHostname + " not available"));
                    }
                } else if (amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_NO_VLANS)) {
                    answeredPermanentFailure = true;
                    return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(2, "Could not reserve vlan tags for " + cmHostname + ""));
                }
            }

            state = State.ALLOCATED;

            assert !amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_NO_VLANS);
            try {
                this.manifest = createManifest(requestRspec);
            } catch (NoVlansAvailableException e) {
                if (e.getVlan() != null) {
                    return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(5, "TEST BUG DETECTED: BUG in tester itself: This vlan not available error should have been handled above: vlan \"" + e.getVlan() + "\""));
                } else
                    //occurs if suggested "any" and no vlans available from the received set of available vlans
                    return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapErrorReply(2, "Could not reserve vlan tags for " + cmHostname + ""));
            }
            List<ComponentInfo> nodes = lookupByType("node");
            ComponentInfo node1 = nodes.get(0);
            String sliverUrn = node1.sliverId;

            List<TestStitchingXmlRpcHelper.SliverInfo> sliverInfos = new ArrayList<>();
            sliverInfos.add(new TestStitchingXmlRpcHelper.SliverInfo("geni_allocated", null, state+"", new Date(System.currentTimeMillis()+(60*60*1000l)), sliverUrn, null));
            return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapSuccessAmv3Reply(manifest, null, sliverInfos));
        }
        if (request.getMethodName().equals("Provision") && request.getParameterCount() >= 1) {
            if (!state.equals(State.ALLOCATED)) throw new RuntimeException("am is in wrong state for "+request.getMethodName()+": "+state);

            state = State.PROVISIONED;

            List<ComponentInfo> nodes = lookupByType("node");
            ComponentInfo node1 = nodes.get(0);
            String sliverUrn = node1.sliverId;

            List<TestStitchingXmlRpcHelper.SliverInfo> sliverInfos = new ArrayList<>();
            sliverInfos.add(new TestStitchingXmlRpcHelper.SliverInfo("geni_provisioned", "geni_configuring", state+"", new Date(System.currentTimeMillis()+(60*60*1000l)), sliverUrn, null));
            return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapSuccessAmv3Reply(this.manifest, null, sliverInfos));
        }
        if (request.getMethodName().equals("PerformOperationalAction") && request.getParameterCount() >= 1) {
            if (!state.equals(State.PROVISIONED)) throw new RuntimeException("am is in wrong state for "+request.getMethodName()+": "+state);

            state = State.STARTING;

            List<ComponentInfo> nodes = lookupByType("node");
            ComponentInfo node1 = nodes.get(0);
            String sliverUrn = node1.sliverId;

            String allStatus = state.equals(State.ALLOCATED) ? "geni_allocated" : "geni_provisioned";
            String opStatus = state.equals(State.ALLOCATED) ? null : (state.equals(State.READY) ? "geni_ready" : "geni_configuring");

            List<TestStitchingXmlRpcHelper.SliverInfo> sliverInfos = new ArrayList<>();
            assert sliverUrn != null;
            sliverInfos.add(new TestStitchingXmlRpcHelper.SliverInfo(allStatus, opStatus, state+"", new Date(System.currentTimeMillis()+(60*60*1000l)), sliverUrn, null));
            return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapSuccessAmv3SliverInfoReply(sliverInfos));
//            return TestStitchingXmlRpcHelper.wrapSuccessXmlRpcReply("<i4>0</i4>");
        }
        if ((request.getMethodName().equals("Status") || request.getMethodName().equals("Renew")) && request.getParameterCount() >= 2) {
            if (state.equals(State.NO)) throw new RuntimeException("am is in wrong state for "+request.getMethodName()+": "+state);

            String error = null;
            if (totalStatusRequestsReceived == 1 && amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.FAIL_ONCE_SECOND_STATUS_BANDWIDTH)) {
                state = State.FAILED;
                error = "Bandwidth PCE(PCE_CREATE_FAILED): 'Unable to find path because the maximum bandwidth of a link has been exceeded.";
            }
            if (totalStatusRequestsReceived == 1 && amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.FAIL_ONCE_SECOND_STATUS_GENERICCAUSE)) {
                state = State.FAILED;
                error = "unspecified error";
            }
            if (totalStatusRequestsReceived == 1 && amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.FAIL_ONCE_SECOND_STATUS_VLAN_UNAVAILABLE)) {
                state = State.FAILED;
                error = "vlan "+activeVlan+" is unavailable";
            }
            if (statusRequestsReceived == 1 && amConfig.hasSpecialReply(StitchingTestSetupConfig.SpecialReply.FAIL_ALLWAYS_SECOND_STATUS_VLAN_UNAVAILABLE)) {
                state = State.FAILED;
                error = "vlan "+activeVlan+" is unavailable";
            }

            // struct Status(string urns[], struct credentials[], struct options)
            String sliceUrn = (String) ((List) request.getParameter(0)).get(0);
            if (state.equals(State.STARTING) && statusRequestsReceived >= statusRequestsBeforeReady)
                state = State.READY;

            statusRequestsReceived += 1;
            totalStatusRequestsReceived += 1;

            List<ComponentInfo> nodes = lookupByType("node");
            ComponentInfo node1 = nodes.get(0);
            String sliverUrn = node1.sliverId;

            String allStatus = state.equals(State.ALLOCATED) ? "geni_allocated" : "geni_provisioned";
            String opStatus;
            switch(state) {
                case ALLOCATED: { opStatus = null; break; }
                case FAILED: { opStatus = "geni_failed"; break; }
                case READY: { opStatus = "geni_ready"; break; }
                default: { opStatus = "geni_configuring"; break; }
            }

            List<TestStitchingXmlRpcHelper.SliverInfo> sliverInfos = new ArrayList<>();
            sliverInfos.add(new TestStitchingXmlRpcHelper.SliverInfo(allStatus, opStatus, state+"", new Date(System.currentTimeMillis()+(60*60*1000l)), sliverUrn, error));
            return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapSuccessAmv3Reply(null, sliceUrn, sliverInfos));
        }
        if (request.getMethodName().equals("Describe") && request.getParameterCount() >= 2) {
            if (state.equals(State.NO)) throw new RuntimeException("am is in wrong state for "+request.getMethodName()+": "+state);

            // struct Describe(string urns[], struct credentials[], struct options[])
            String sliceUrn = (String) ((List) request.getParameter(0)).get(0);

            List<ComponentInfo> nodes = lookupByType("node");
            ComponentInfo node1 = nodes.get(0);
            String sliverUrn = node1.sliverId;

            String allStatus = state.equals(State.ALLOCATED) ? "geni_allocated" : "geni_provisioned";
            String opStatus = state.equals(State.ALLOCATED) ? null : (state.equals(State.READY) ? "geni_ready" : "geni_configuring");

            List<TestStitchingXmlRpcHelper.SliverInfo> sliverInfos = new ArrayList<>();
            sliverInfos.add(new TestStitchingXmlRpcHelper.SliverInfo(allStatus, opStatus, state+"", new Date(System.currentTimeMillis()+(60*60*1000l)), sliverUrn, null));
            return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapSuccessAmv3Reply(this.manifest, sliceUrn, sliverInfos));
        }
        if (request.getMethodName().equals("Delete") && request.getParameterCount() > 0) {
            state = State.NO;

            activeVlan = null;

            seenRequests.clear();
            sawAnyRequest = false;
            statusRequestsReceived = 0;

            List<TestStitchingXmlRpcHelper.SliverInfo> sliverInfos = new ArrayList<>();

            List<ComponentInfo> nodes = lookupByType("node");
            if (!nodes.isEmpty()) {
                ComponentInfo node1 = nodes.get(0);
                String sliverUrn = node1.sliverId;
                sliverInfos.add(new TestStitchingXmlRpcHelper.SliverInfo("geni_unallocated", null, state+"", new Date(System.currentTimeMillis() + (60 * 60 * 1000l)), sliverUrn, null));
            }

            return new MockSfaConnection.MockXmlRpcAnswer(TestStitchingXmlRpcHelper.wrapSuccessAmv3SliverInfoReply(sliverInfos));
        }
        return null;
    }
}
